@@ -17,8 +17,9 @@ import re
 import gc
 import math as m
 import time
+
 # visualization imports
-from bokeh.plotting import figure, output_file, show, output_notebook
+from bokeh.plotting import figure, output_file, show, output_notebook, save
 from bokeh.models import ColumnDataSource, HoverTool, Div, FactorRange
 from bokeh.layouts import gridplot, column, row
 from bokeh.palettes import Category10, Category20, Colorblind
@@ -431,7 +432,7 @@ class tfr_dataset:
             # extract rows for these sample_ids in parallel
             df = data[self.col_list].query("{}==@sample_id".format(self.id_col))
             arr, pad_arr = self.static_arrs(df, use_memmap=False, prefix='fill_buffer') 
-            valid_indices = np.where(np.count_nonzero(arr[:,:self.window_len-self.fh,self.target_index].astype(np.float32)>0,axis=1)>=self.min_nz)
+            valid_indices = np.where(np.count_nonzero(arr[:,:self.window_len-self.fh,self.target_index].astype(np.float32)>=0,axis=1)>=self.min_nz)
             arr = arr[valid_indices]
             pad_arr = pad_arr[valid_indices]
             m_in, m_out, s, w  = self.preprocess(arr, pad_arr)
@@ -442,35 +443,15 @@ class tfr_dataset:
             scale.append(s)
             weights.append(w)
             
-        model_in = np.vstack(model_in)
-        model_out = np.vstack(model_out)
-        scale = np.vstack(scale)
-        weights = np.vstack(weights)    
-        
-        '''
-        data_gen = self.data_generator(data)
-        
-        model_in = []
-        model_out =[]
-        scale = []
-        weights =[]
-        
-        obs_count = 0
-        for i, (m_in, m_out, s, w) in enumerate(data_gen):
-            obs_count += m_in.shape[0]
-            if obs_count < num_observations: 
-                model_in.append(m_in)
-                model_out.append(m_out)
-                scale.append(s)
-                weights.append(w)
-            else:
-                break
-                
         model_in = np.vstack(model_in)
         model_out = np.vstack(model_out)
         scale = np.vstack(scale)
         weights = np.vstack(weights)
-        '''
+
+        # shuffle
+        p = np.random.permutation(len(model_in))
+        model_in, model_out, scale, weights = model_in[p], model_out[p], scale[p], weights[p]
+
         return model_in.astype(str), model_out.astype(np.float32), scale.astype(np.float32), weights.astype(np.float32)
         
             
@@ -675,8 +656,8 @@ class tfr_dataset:
         model_in = np.concatenate((model_in, mask), axis=-1)
 
         # sample weights
-        weights = np.squeeze(target_nz_mean)/np.quantile(np.squeeze(target_nz_mean), q=0.8)
-        weights = np.clip(weights, a_min=0.5, a_max=1.0)
+        weights = np.around(np.log10(np.squeeze(target_nz_mean) + 10),2) #/ np.quantile(np.squeeze(target_nz_mean), q=0.8)
+        weights = np.clip(weights, a_min=1.0, a_max=2.0)
         weights = weights.reshape(-1,1)
         #weights = np.expand_dims(weights.reshape(-1,1), axis=-1)
         #weights = np.tile(weights, [1,sequence_len,1])
@@ -791,31 +772,37 @@ class tfr_dataset:
             model_in_y, model_out_y, scale_y, weights_y = self.static_dataset(test_data, batchsize=self.batch*max(m.ceil(test_data[self.id_col].nunique()*0.01), 200), use_memmap=use_memmap, prefix='test')
             
             if tf.__version__ < "2.7.0":
-                trainset = tf.data.Dataset.from_tensor_slices((model_in_x, model_out_x, scale_x, weights_x)).shuffle(SHUFFLE_BUFFER_SIZE, reshuffle_each_iteration=True).batch(self.batch, drop_remainder=True)
-                testset = tf.data.Dataset.from_tensor_slices((model_in_y, model_out_y, scale_y, weights_y)).batch(self.batch, drop_remainder=False)
+                trainset = tf.data.Dataset.from_tensor_slices((model_in_x, model_out_x, scale_x, weights_x)).batch(self.batch, drop_remainder=True)
+                testset = tf.data.Dataset.from_tensor_slices((model_in_y, model_out_y, scale_y, weights_y)).batch(self.batch, drop_remainder=True)
             else:
-                trainset = tf.data.Dataset.from_tensor_slices((model_in_x, model_out_x, scale_x, weights_x)).shuffle(SHUFFLE_BUFFER_SIZE, reshuffle_each_iteration=True).batch(self.batch, drop_remainder=True, num_parallel_calls=self.PARALLEL_DATA_JOBS)
-                testset = tf.data.Dataset.from_tensor_slices((model_in_y, model_out_y, scale_y, weights_y)).batch(self.batch, drop_remainder=False, num_parallel_calls=self.PARALLEL_DATA_JOBS)
+                trainset = tf.data.Dataset.from_tensor_slices((model_in_x, model_out_x, scale_x, weights_x)).batch(self.batch, drop_remainder=True, num_parallel_calls=self.PARALLEL_DATA_JOBS)
+                testset = tf.data.Dataset.from_tensor_slices((model_in_y, model_out_y, scale_y, weights_y)).batch(self.batch, drop_remainder=True, num_parallel_calls=self.PARALLEL_DATA_JOBS)
         
-        else:
+        elif fill_buffer:
             #avg_train_samples = max(1, int((train_data.groupby(self.id_col).size().mean() - self.window_len)/self.interleave))
             #num_train_observations = avg_train_samples*train_data[self.id_col].nunique()
             
             #avg_test_samples = max(1, int((test_data.groupby(self.id_col).size().mean() - self.window_len)))
             #num_test_observations = avg_test_samples*test_data[self.id_col].nunique()
             
-            model_in_x, model_out_x, scale_x, weights_x = self.fill_buffer(train_data) # , num_observations=num_train_observations)
-            model_in_y, model_out_y, scale_y, weights_y = self.fill_buffer(test_data) # , num_observations=num_test_observations)
-            
+            model_in_x, model_out_x, scale_x, _ = self.fill_buffer(train_data) # , num_observations=num_train_observations)
+            model_in_y, model_out_y, scale_y, _ = self.fill_buffer(test_data) # , num_observations=num_test_observations)
+
+            # weights over entire dataset based on scale
+            weights_x = np.around(np.log10(np.squeeze(scale_x[:, -1:, :]) + 10),2)
+            weights_x = weights_x.reshape(-1, 1)
+            weights_y = np.around(np.log10(np.squeeze(scale_y[:, -1:, :]) + 10),2)
+            weights_y = weights_y.reshape(-1, 1)
+
             TRAIN_SHUFFLE_BUFFER_SIZE = model_in_x.shape[0] # num_train_observations
             TEST_SHUFFLE_BUFFER_SIZE = model_in_y.shape[0] # num_test_observations
             
             if tf.__version__ < "2.7.0":
-                trainset = tf.data.Dataset.from_tensor_slices((model_in_x, model_out_x, scale_x, weights_x)).shuffle(TRAIN_SHUFFLE_BUFFER_SIZE, reshuffle_each_iteration=True).batch(self.train_test_batch_size, drop_remainder=True)
-                testset = tf.data.Dataset.from_tensor_slices((model_in_y, model_out_y, scale_y, weights_y)).shuffle(TEST_SHUFFLE_BUFFER_SIZE, reshuffle_each_iteration=True).batch(self.train_test_batch_size, drop_remainder=False)
+                trainset = tf.data.Dataset.from_tensor_slices((model_in_x, model_out_x, scale_x, weights_x)).batch(self.train_test_batch_size, drop_remainder=True)
+                testset = tf.data.Dataset.from_tensor_slices((model_in_y, model_out_y, scale_y, weights_y)).batch(self.train_test_batch_size, drop_remainder=True)
             else:
-                trainset = tf.data.Dataset.from_tensor_slices((model_in_x, model_out_x, scale_x, weights_x)).shuffle(TRAIN_SHUFFLE_BUFFER_SIZE, reshuffle_each_iteration=True).batch(self.train_test_batch_size, drop_remainder=True, num_parallel_calls=self.PARALLEL_DATA_JOBS)
-                testset = tf.data.Dataset.from_tensor_slices((model_in_y, model_out_y, scale_y, weights_y)).shuffle(TEST_SHUFFLE_BUFFER_SIZE, reshuffle_each_iteration=True).batch(self.train_test_batch_size, drop_remainder=False, num_parallel_calls=self.PARALLEL_DATA_JOBS)
+                trainset = tf.data.Dataset.from_tensor_slices((model_in_x, model_out_x, scale_x, weights_x)).batch(self.train_test_batch_size, drop_remainder=True, num_parallel_calls=self.PARALLEL_DATA_JOBS)
+                testset = tf.data.Dataset.from_tensor_slices((model_in_y, model_out_y, scale_y, weights_y)).batch(self.train_test_batch_size, drop_remainder=True, num_parallel_calls=self.PARALLEL_DATA_JOBS)
          
         return trainset, testset
         
@@ -903,7 +890,28 @@ class tfr_dataset:
         
         return [input_tensor, scale, id_arr, date_arr]
     
-    def show_ts_samples(self, data, sample_ids=[], n_samples=10, n_col=2, plot_size=(300,600), save=False, filename='ts_samples.html'):
+    def acf(self, series):
+        # https://stackoverflow.com/questions/43344406/is-there-a-bokeh-version-of-pandas-autocorrelation-plot-method
+        n = len(series)
+        data = np.asarray(series)
+        mean = np.mean(data)
+        c0 = np.sum((data - mean) ** 2) / float(n)
+        def r(h):
+            acf_lag = ((data[:n - h] - mean) * (data[h:] - mean)).sum() / float(n) / c0
+            return round(acf_lag, 3)
+        x = np.arange(n) # Avoiding lag 0 calculation
+        acf_coeffs = pd.Series(map(r, x)).round(decimals = 3)
+        acf_coeffs = acf_coeffs + 0
+        return acf_coeffs
+
+    def significance(self, series):
+        # https://stackoverflow.com/questions/43344406/is-there-a-bokeh-version-of-pandas-autocorrelation-plot-method
+        n = len(series)
+        z95 = 1.959963984540054 / np.sqrt(n)
+        z99 = 2.5758293035489004 / np.sqrt(n)
+        return (z95,z99)
+    
+    def show_ts_samples(self, data, sample_ids=[], n_samples=10, n_col=2, plot_size=(300,600), save=True, filename='ts_samples.html'):
         
         # sort dataset
         data = self.sort_dataset(data)
@@ -948,11 +956,25 @@ class tfr_dataset:
             plots = []
             for col in num_cols:
                 y_vals = df_sample[col].unique().tolist()
-                p = figure( plot_height=h, plot_width=w, tools=TOOLS, x_axis_label='timestep', y_axis_label=col, title = "{}_{}".format(sid,col))
+                p = figure( plot_height=h, plot_width=w, tools=TOOLS, x_axis_label='timestep', y_axis_label=col, title = "{} {}".format(sid,col))
                 p.line(x='index', y=col, source=source)
                 tooltips = [(col,'@{}'.format(col)), (date_col,'@{}'.format(date_col))]
                 p.add_tools(HoverTool(tooltips=tooltips))
                 plots.append(p)
+            
+            # autocorr plot
+            series = df_sample[target_col]
+            x = pd.Series(range(1, len(series)+1), dtype = float)
+            z95, z99 = self.significance(series)
+            y = self.acf(series)
+            p = figure(title="{} Auto-Correlation Plot".format(sid), plot_height=h, plot_width=w, x_axis_label="Lag", y_axis_label="Autocorrelation")
+            p.line(x, z99, line_dash='dashed', line_color='grey')
+            p.line(x, z95, line_color = 'grey')
+            p.line(x, y=0.0, line_color='black')
+            p.line(x, z99*-1, line_dash='dashed', line_color='grey')
+            p.line(x, z95*-1, line_color = 'grey')
+            p.line(x, y, line_width=2)
+            plots.append(p)
     
             for i in range(len(plots)):
                 if i==0:
@@ -962,14 +984,15 @@ class tfr_dataset:
         
             subplots = [plots[i:i + n_col] for i in range(0, len(plots), n_col)]
             saveplots += subplots
-           
+            
             if len(cat_covar_cols)>0:
                 plots = []
                 for col in cat_covar_cols:
                     df_cat = df_sample.groupby(col).agg({target_col:'mean'}).reset_index()
                     source = ColumnDataSource(data=df_cat)
-                    p = figure(x_range=df_cat[col].astype(str).unique(), plot_height=h, plot_width=w, tools=TOOLS, x_axis_label='timestep', y_axis_label=col, title = "{}_{}".format(sid,col))
+                    p = figure(x_range=df_cat[col].astype(str).unique(), plot_height=h, plot_width=w, tools=TOOLS, x_axis_label=col, y_axis_label='Mean ' + str(target_col), title = "{} {}".format(sid,col))
                     p.vbar(x=col, top=target_col, source=source, width=0.8)
+                    p.xaxis.major_label_orientation = "vertical"
                     tooltips = [(col,'@{}'.format(col)),(target_col,'@{}'.format(target_col))]
                     p.add_tools(HoverTool(tooltips=tooltips))
                     plots.append(p)
@@ -981,11 +1004,11 @@ class tfr_dataset:
             else:
                 html = """<h3>{}: Time Series Plots with Numerical Covariates</h3>""".format(target_col)
                 sup_title = Div(text=html)
-        
+                
         grid = gridplot(saveplots)
         show(column(sup_title, grid))
         
-    def show_processed_ts_samples(self, data, n_samples=10, n_col=2, plot_size=(300,400)):
+    def show_processed_ts_samples(self, data, n_samples=10, n_col=2, plot_size=(300,400), save=True, filename='processed_ts_samples.html'):
         # sort
         data = self.sort_dataset(data)
         
@@ -1006,7 +1029,11 @@ class tfr_dataset:
         h,w = plot_size
         x_range = np.arange(self.window_len).tolist() 
         
+        if save:
+            output_file(filename)
+            
         # display sample series
+        layouts = []
         for i in range(len(x)):
             
             if i >= n_samples:
@@ -1054,8 +1081,12 @@ class tfr_dataset:
             html = """<h3>{}: Scaled Time Series Plots (Numerical Columns)</h3>""".format(sid)
             sup_title = Div(text=html)
             grid = gridplot(subplots)
-            show(column(sup_title, grid))
+            layout = column(sup_title, grid, sizing_mode="stretch_both")
+            layouts.append(layout)
             i += 1
+        
+        supergrid = gridplot(layouts, ncols=1)
+        show(supergrid)
 
 
 
